"""
Copyright (c) 2021 James Korte, Zachary Chin

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
following conditions are met:
1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer
   in the documentation and/or other materials provided with the distribution.
3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

--------------------------------------------------------------------------------
Change Log:
--------------------------------------------------------------------------------
  21-June-2021  : (Zachary Chin, James Korte) : Initial code from masters project
02-August-2021  :               (James Korte) : Updated for MR-BIAS code v0.0
  23-June-2022  :               (James Korte) : GitHub Release   MR-BIAS v1.0
"""

import os
from enum import IntEnum
from collections import OrderedDict
from abc import ABC, abstractmethod

import SimpleITK as sitk
import yaml
import numpy as np
import matplotlib.pyplot as plt

# for pdf output
from reportlab.lib.pagesizes import landscape
from reportlab.pdfgen import canvas
from reportlab.lib.utils import ImageReader


# Code to handle running each module as a test case (from within the module)
from pathlib import Path
import sys
file = Path(__file__).resolve()
parent, root = file.parent, file.parents[1]
if str(root) not in sys.path:
    sys.path.insert(1, str(root))
# import required mrbias modules
from mrbias import misc_utils as mu
from mrbias.misc_utils import LogLevels
from mrbias.misc_utils import ROI_IDX_LABEL_MAP, T1_ROI_LABEL_IDX_MAP, T2_ROI_LABEL_IDX_MAP
import mrbias.scan_session as scan_session
import mrbias.image_sets as image_sets


# for future expansion / different phantoms
class ROITypeOptions(IntEnum):
    SPHERE = 1

class RegistrationOptions(IntEnum):
    COREL_GRADIENTDESCENT = 1
    MSME_GRIDSEARCH = 2
    TWOSTAGE_MSMEGS_CORELGD = 3


def main():
    # setup the logger to write to file
    mu.initialise_logger("roi_detect.log", force_overwrite=True, write_to_screen=True)
    # setup a pdf to test the pdf reporting
    pdf = mu.PDFSettings()
    c = canvas.Canvas("roi_detect.pdf", landscape(pdf.page_size))

    # target images to test (skyra)
    dcm_dir_a = os.path.join(mu.reference_data_directory(), "mrbias_testset_B")
    ss = scan_session.ScanSessionSiemensSkyra(dcm_dir_a)
    test_geometric_images = ss.get_geometric_images()
    test_hard = test_geometric_images[0]
    test_easy = test_geometric_images[1]
    test_geo_vec = [test_hard, test_easy]
    case_name_vec = ["large miss-alignment", "small miss-alignment"]

    # get the T1 and T2 imagesets
    t1_vir_imagesets = ss.get_t1_vir_image_sets()
    t1_vfa_imagesets = ss.get_t1_vfa_image_sets()
    t2_mse_imagesets = ss.get_t2_mse_image_sets()

    # set the test template
    TEST_TEMPLATE_DIR = os.path.join(mu.reference_template_directory(), "siemens_skyra_3p0T")

    for test_target_im, case_str in zip(test_geo_vec, case_name_vec):
        # create a roi detector
        roi_detector = ROIDetector(test_target_im, TEST_TEMPLATE_DIR)
        # detect the ROIs and return the masks on the target image
        roi_detector.detect()

        # output to pdf
        figure_title = "ROI DETECTOR : %s" % case_str
        roi_detector.write_pdf_summary_page(c, figure_title)

    # for each of the image datasets show the detected ROIs
    for t1_vir_imageset in t1_vir_imagesets:
        t1_vir_imageset.update_ROI_mask()  # trigger a mask update
        t1_vir_imageset.write_roi_pdf_page(c)
    for t1_vfa_imageset in t1_vfa_imagesets:
        t1_vfa_imageset.update_ROI_mask()  # trigger a mask update
        t1_vfa_imageset.write_roi_pdf_page(c)
    for t2_mse_imageset in t2_mse_imagesets:
        t2_mse_imageset.update_ROI_mask()  # trigger a mask update
        t2_mse_imageset.write_roi_pdf_page(c)

    # save the pdf report
    c.save()
    mu.log("------ FIN -------", LogLevels.LOG_INFO)
    plt.show()


class ROITemplate(object):
    def __init__(self, template_dir,
                 dcm_subdir="dicom",
                 t1_rois_file="default_T1_rois.yaml",
                 t2_rois_file="default_T2_rois.yaml"):
        dcm_dir = os.path.join(template_dir, dcm_subdir)
        t1_yaml_file = os.path.join(template_dir, t1_rois_file)
        t2_yaml_file = os.path.join(template_dir, t2_rois_file)
        if not os.path.isdir(dcm_dir):
            mu.log("ROITemplate::__init__(): invalid dicom dir : %s" % dcm_dir, LogLevels.LOG_WARNING)
        if not os.path.isfile(t1_yaml_file):
            mu.log("ROITemplate::__init__(): invalid t1 roi yaml file : %s" % t1_yaml_file, LogLevels.LOG_WARNING)
        if not os.path.isfile(t2_yaml_file):
            mu.log("ROITemplate::__init__(): invalid t2 roi yaml file : %s" % t2_yaml_file, LogLevels.LOG_WARNING)
        # load the image file
        mu.log("ROITemplate::init(): loading template geometry image from DCM dir: %s" %
               dcm_dir, LogLevels.LOG_INFO)
        self.image = mu.get_sitk_image_from_dicom_image_folder(dcm_dir)
        # parse the ROI yaml files
        self.t1_roi_dict = self.parse_t1_yaml(t1_yaml_file)
        self.t2_roi_dict = self.parse_t2_yaml(t2_yaml_file)

    def parse_t1_yaml(self, yaml_file):
        return self.__parse_roi_yaml_file(yaml_file, T1_ROI_LABEL_IDX_MAP)
    def parse_t2_yaml(self, yaml_file):
        return self.__parse_roi_yaml_file(yaml_file, T2_ROI_LABEL_IDX_MAP)
    def __parse_roi_yaml_file(self, yaml_file, roi_label_idx_map):
        roi_dict = OrderedDict()
        in_dict = yaml.full_load(open(yaml_file))
        for roi_label, roi_dx in roi_label_idx_map.items():
            if roi_label in in_dict.keys():
                # found in yaml file
                yaml_roi = in_dict[roi_label]
                if "roi_type" in yaml_roi.keys():
                    roi_type = yaml_roi["roi_type"]
                    if roi_type == "sphere":
                        # check all sphere fields are available and create Spherical ROI
                        for field in ['roi_radius_mm', 'ctr_vox_coords']:
                            if not (field in yaml_roi.keys()):
                                mu.log("ROITemplate::__parse_roi_yaml_file(): skipping ROI(%s) no expected field '%s' "
                                       "specified in yaml file : %s" % (roi_label, field, yaml_file), LogLevels.LOG_WARNING)
                        roi_radius_mm = yaml_roi["roi_radius_mm"]
                        ctr_vox_coords = yaml_roi["ctr_vox_coords"]
                        assert isinstance(roi_radius_mm, float), "ROITemplate::__parse_roi_yaml_file(): " \
                                                                 "roi_radius_mm expected datatype is float (not %s)" % \
                                                                 type(roi_radius_mm)
                        assert isinstance(ctr_vox_coords, list), "ROITemplate::__parse_roi_yaml_file(): " \
                                                                 "ctr_vox_coords expected datatype is list (not %s)" % \
                                                                 type(ctr_vox_coords)
                        roi_dict[roi_dx] = ROISphere(roi_label, roi_dx, ctr_vox_coords, roi_radius_mm)
                else:
                    mu.log("ROITemplate::__parse_roi_yaml_file(): skipping ROI(%s) no field 'roi_type' "
                           "specified in yaml file : %s" % (roi_label, yaml_file), LogLevels.LOG_WARNING)
            else:
                mu.log("ROITemplate::__parse_roi_yaml_file(): ROI(%s) not specified in yaml file : %s" %
                       (roi_label, yaml_file), LogLevels.LOG_WARNING)
        # return the ROI dictionary
        return roi_dict

    def get_T1_mask_image(self):
        return self.__get_mask_image(self.t1_roi_dict)
    def get_T2_mask_image(self):
        return self.__get_mask_image(self.t2_roi_dict)
    def __get_mask_image(self, roi_dict):
        """
        Create a template mask image of the same image grid as the template image
        Args:
            roi_dict (OrderedDict): with a roi_idx key and values of type ROI (only supports ROISphere)
        Returns:
            SimpleITK.Image: a template mask with 0 as background and ROIs with values as defined in points
        """
        # Create a numpy image array of zeros with the same size as the geometric image
        fixed_geo_arr = sitk.GetArrayFromImage(self.image)
        fixed_geo_spacing = np.array(self.image.GetSpacing())
        mask_arr = np.zeros_like(fixed_geo_arr)
        for roi_dx, roi in roi_dict.items():
            # rely on the concrete class to draw its own ROI on the mask image
            roi.draw(mask_arr, fixed_geo_spacing)
        # Convert the roi mask array to a simpleITK image with matched spatial properties to the
        # fixed geometric image
        masked_image = sitk.GetImageFromArray(mask_arr)
        masked_image.SetOrigin(self.image.GetOrigin())
        masked_image.SetDirection(self.image.GetDirection())
        masked_image.SetSpacing(self.image.GetSpacing())
        return masked_image

    def get_t1_slice_dx(self):
        slice_vec = []
        for roi in self.t1_roi_dict.values():
            slice_vec.append(roi.get_slice_dx())
        return int(np.median(np.array(slice_vec)))

    def get_t2_slice_dx(self):
        slice_vec = []
        for roi in self.t2_roi_dict.values():
            slice_vec.append(roi.get_slice_dx())
        return int(np.median(np.array(slice_vec)))

    def get_t1_roi_values(self):
        return list(T1_ROI_LABEL_IDX_MAP.values())
    def get_t2_roi_values(self):
        return list(T2_ROI_LABEL_IDX_MAP.values())



class ROI(ABC):
    def __init__(self, label, roi_index):
        self.label = label
        self.roi_index = roi_index
        assert self.roi_index in ROI_IDX_LABEL_MAP.keys(), "ROI::__init__: roi index is invalid, idx=%d" % self.roi_index

    """
    Draw the ROI into the passes mask array, marking the ROI with the global ROI index
    """
    @abstractmethod
    def draw(self, arr, spacing):
        return None

    @abstractmethod
    def get_slice_dx(self):
        return None

class ROISphere(ROI):
    def __init__(self, label, roi_index,
                 ctr_vox_coords, radius_mm):
        super().__init__(label, roi_index)
        self.ctr_vox_coords = ctr_vox_coords
        self.radius_mm = radius_mm
        mu.log("\t\tROISphere::__init__(): sphere (%d : %s) created!" % (roi_index, label), LogLevels.LOG_INFO)

    def draw(self, arr, spacing):
        # calculate how many voxels to achieve the radius
        radius_vox = self.radius_mm / spacing
        z, y, x = np.ogrid[:arr.shape[2],:arr.shape[1],:arr.shape[0]]
        # Assigns the masked pixels in the copy image array to corresponding pixel values
        # TODO: validate ordering of radius_vox [0,1,2] with an image which is less isotropic
        distance_from_centre = np.sqrt(((x - self.ctr_vox_coords[0]) / radius_vox[0]) ** 2 +
                                       ((y - self.ctr_vox_coords[1]) / radius_vox[1]) ** 2 +
                                       ((z - self.ctr_vox_coords[2]) / radius_vox[2]) ** 2)
        sphere_mask = distance_from_centre <= 1.0
        arr[sphere_mask] = self.roi_index

    def get_slice_dx(self):
        return self.ctr_vox_coords[2]


class ROIDetector(object):
    """
    A class for detecting regions of interest (ROIs) on a 3D T1 weighted image of the ISMRM/NIST qMRI System Phantom.

    The target image is registered to a template image which has associated template ROIs. The ROIs are then warped
    from the template image space onto the target image space.

    Attributes:
        target_geo_im (ImageSet.ImageGeometric): the target image on which we want to detect ROIs
        reg_method (RegistrationMethodAbstract): the registration method to align the target and template images
        fixed_geom_im (SimpleITK.Image): the template T1 weighted 3D MRI image
        transform (SimpleITK.Transform): the estimated transform between the target image and template image
    """

    def __init__(self,
                 target_geometry_image,
                 template_directory,
                 registration_method=RegistrationOptions.TWOSTAGE_MSMEGS_CORELGD):
        """
        Class constructor stores the target image, registration method choice and loads the template image.

        Args:
            target_geometry_image (ImageSet.ImageGeometric): the target image on which we want to detect ROIs
            template_directory (directory): a template directory with a DICOM folder and ROI yaml files
            registration_method (RegistrationOptions): a flag representing the registration method to use
        """
        assert isinstance(target_geometry_image, image_sets.ImageGeometric), \
            "ROIDetector::init(): taget_geometry_image is expected as a SimpleITK image (not %s)" % \
            type(target_geometry_image)
        self.target_geo_im = target_geometry_image
        self.reg_method = registration_method
        self.roi_template = ROITemplate(template_directory)
        self.fixed_geom_im = self.roi_template.image
        assert isinstance(self.fixed_geom_im, sitk.Image), \
            "ROIDetector::init(): self.fixed_geom_im (template image) is expected as a SimpleITK image (not %s)" % \
            type(self.fixed_geom_im)
        self.transform = None

    def detect(self):
        """ Detect the ROIs on the target image, by registering the target image to template image."""
        # create relevant registration class
        # rego = None
        # if self.reg_method == RegistrationOptions.COREL_GRADIENTDESCENT:
        #     rego = RegistrationCorrelationGradientDescent(self.fixed_geom_im, self.target_geo_im)
        # elif self.reg_method == RegistrationOptions.MSME_GRIDSEARCH:
        #     rego = RegistrationMSMEGridSearch(self.fixed_geom_im, self.target_geo_im)
        # assert rego is not None, "ROIDetector::detect(): invalid registration method, %s" % str(self.reg_method)
        rego = RegistrationMethodAbstract.generate_registration_instance(self.reg_method,
                                                                         self.fixed_geom_im,
                                                                         self.target_geo_im.get_image())

        mu.log("ROIDetector::detect():",
               LogLevels.LOG_INFO)
        mu.log("\t Iteration |   Metric   |   Optimizer Position",
               LogLevels.LOG_INFO)
        transform, metric = rego.register()
        # store the transform for warping ROIs
        self.transform = transform
        # store the resulting masks on the geometric image
        self.target_geo_im.set_T1_mask(self.get_detected_T1_mask())
        self.target_geo_im.set_T2_mask(self.get_detected_T2_mask())

    def get_detected_T1_mask(self):
        """
        Returns:
            SimpleITK.Image: T1 mask in target image space, with 0 for background and 1-14 for detected ROIs
        """
        mu.log("ROIDetector::get_detected_T1_mask()", LogLevels.LOG_INFO)
        fixed_t1_mask_im = self.get_fixed_T1_mask()
        return self.__get_registered_mask(fixed_t1_mask_im, self.target_geo_im.get_image())

    def get_detected_T2_mask(self):
        """
        Returns:
            SimpleITK.Image: T2 mask  in target image space, with 0 for background and 15-28 for detected ROIs
        """
        mu.log("ROIDetector::get_detected_T2_mask()", LogLevels.LOG_INFO)
        fixed_t2_mask_im = self.get_fixed_T2_mask()
        return self.__get_registered_mask(fixed_t2_mask_im, self.target_geo_im.get_image())

    def get_fixed_T1_mask(self):
        """
        Returns:
            SimpleITK.Image: T1 mask in template image space, with 0 for background and 1-14 for detected ROIs
        """
        mu.log("ROIDetector::get_fixed_T1_mask()", LogLevels.LOG_INFO)
        return self.roi_template.get_T1_mask_image()

    def get_fixed_T2_mask(self):
        """
        Returns:
            SimpleITK.Image: T2 mask  in template image space, with 0 for background and 15-28 for detected ROIs
        """
        mu.log("ROIDetector::get_fixed_T2_mask()", LogLevels.LOG_INFO)
        return self.roi_template.get_T2_mask_image()

    def visualise_fixed_T1_rois(self, ax=None):
        geo_arr = sitk.GetArrayFromImage(self.fixed_geom_im)
        roi_arr = sitk.GetArrayFromImage(self.get_fixed_T1_mask())
        t1_slice_dx = self.roi_template.get_t1_slice_dx()
        t1_roi_values = self.roi_template.get_t1_roi_values()
        self.__visualise_rois(geo_arr, roi_arr, t1_slice_dx, t1_roi_values, ax,
                              title="T1 (template)")

    def visualise_fixed_T2_rois(self, ax=None):
        geo_arr = sitk.GetArrayFromImage(self.fixed_geom_im)
        roi_arr = sitk.GetArrayFromImage(self.get_fixed_T2_mask())
        t2_slice_dx = self.roi_template.get_t2_slice_dx()
        t2_roi_values = self.roi_template.get_t2_roi_values()
        self.__visualise_rois(geo_arr, roi_arr, t2_slice_dx, t2_roi_values, ax,
                              title="T2 (template)")

    def visualise_detected_T1_rois(self, ax=None):
        self.__visualise_transformed_rois(self.get_detected_T1_mask(),
                                          self.roi_template.get_t1_slice_dx(),
                                          self.roi_template.get_t1_roi_values(), ax,
                                          title="T1 (detected)")

    def visualise_detected_T2_rois(self, ax=None):
        self.__visualise_transformed_rois(self.get_detected_T2_mask(),
                                          self.roi_template.get_t2_slice_dx(),
                                          self.roi_template.get_t2_roi_values(), ax,
                                          title="T2 (detected)")

    def visualise_T1_registration(self, pre_reg_ax=None, post_reg_ax=None, invert_moving=True):
        self.__visualise_registration(self.roi_template.get_t1_slice_dx(),
                                      pre_reg_ax=pre_reg_ax, post_reg_ax=post_reg_ax,
                                      invert_moving=invert_moving,
                                      title="T1")

    def visualise_T2_registration(self, pre_reg_ax=None, post_reg_ax=None, invert_moving=True):
        self.__visualise_registration(self.roi_template.get_t2_slice_dx(),
                                      pre_reg_ax=pre_reg_ax, post_reg_ax=post_reg_ax,
                                      invert_moving=invert_moving,
                                      title="T2")

    # same output as log but to pdf
    def write_pdf_summary_page(self, c, sup_title="ROI Detection: Summary"):
        table_width = 170
        pdf = mu.PDFSettings()
        c.setFont(pdf.font_name, pdf.small_font_size)  # set to a fixed width font

        # draw the summary figure
        # -----------------------------------------------------------
        # setup figure
        f, ((ax1, ax2, ax3, ax4), (ax5, ax6, ax7, ax8)) = plt.subplots(2, 4)
        if sup_title is not None:
            f.suptitle(sup_title)
        f.set_size_inches(14, 6)
        # draw the template rois on the template image
        self.visualise_fixed_T1_rois(ax=ax1)
        self.visualise_fixed_T2_rois(ax=ax5)
        # draw the registration (pre/post)
        self.visualise_T1_registration(invert_moving=False, pre_reg_ax=ax2, post_reg_ax=ax3)
        self.visualise_T2_registration(invert_moving=False, pre_reg_ax=ax6, post_reg_ax=ax7)
        # visualise the transfromed ROI masks on the target image
        self.visualise_detected_T1_rois(ax=ax4)
        self.visualise_detected_T2_rois(ax=ax8)

        # draw it on the pdf
        pil_f = mu.mplcanvas_to_pil(f)
        width, height = pil_f.size
        height_3d, width_3d = pdf.page_width * (height / width), pdf.page_width
        c.drawImage(ImageReader(pil_f),
                    0,
                    pdf.page_height - pdf.top_margin - height_3d - pdf.line_width,
                    width_3d,
                    height_3d)
        plt.close(f)

        c.showPage()  # new page

    def __visualise_registration(self, slice_dx, pre_reg_ax=None, post_reg_ax=None, invert_moving=True, title=None):
        """
        Visualise the image registration with a checkerboard of the fixed and moving images. This function creates
        two checkerboard comparisons; the first with the unregistered target image and the second with the registered
        target image.

        Args:
            slice_dx (int): index of slice to display
            pre_reg_ax (matplotlib.axes): axes to plot the target image prior to registration
            post_reg_ax (matplotlib.axes): axes to plot the target image after registration
            invert_moving (bool): invert the image intensity of the moving image
            title (string): a title for the subplots
        """
        # fake a target geo image to be on same spatial grid as fixed_geo_im for checkerboard filter
        target_geo_im = sitk.GetImageFromArray(sitk.GetArrayFromImage(self.target_geo_im.get_image()))
        target_geo_im.SetOrigin(self.fixed_geom_im.GetOrigin())
        target_geo_im.SetDirection(self.fixed_geom_im.GetDirection())
        target_geo_im.SetSpacing(self.fixed_geom_im.GetSpacing())

        # define intensity limits for visualisation
        t1_im_slice = sitk.GetArrayFromImage(target_geo_im)[slice_dx, :, :]
        vmin = np.mean(t1_im_slice) - 1.0 * np.std(t1_im_slice)
        vmax = np.mean(t1_im_slice) + 2.0 * np.std(t1_im_slice)
        inversion_max = np.mean(t1_im_slice) + 2.5 * np.std(t1_im_slice)

        # pre-registration
        if invert_moving:
            target_geo_im = sitk.InvertIntensity(target_geo_im, maximum=inversion_max)
        target_geo_im_match_type = sitk.Cast(target_geo_im, sitk.sitkUInt16)
        checker_im = sitk.CheckerBoard(self.fixed_geom_im, target_geo_im_match_type, [4, 4, 1])
        checker_arr = sitk.GetArrayFromImage(checker_im)

        # post-registration
        rotated_fixed_im = self.__get_rotated_sampling_im()
        # resample the target image and mask - onto the deformed fixed space
        # (so we can use the reference ROI slice for visualisation)
        target_im = sitk.Resample(self.target_geo_im.get_image(),
                                  rotated_fixed_im,
                                  sitk.Euler3DTransform(), sitk.sitkLinear)
        if invert_moving:
            target_im = sitk.InvertIntensity(target_im, maximum=inversion_max)

        target_im_match_type = sitk.Cast(target_im, sitk.sitkUInt16)
        checker_im_reg = sitk.CheckerBoard(rotated_fixed_im, target_im_match_type, [4, 4, 1])
        checker_arr_reg = sitk.GetArrayFromImage(checker_im_reg)
        if (pre_reg_ax is None) or (post_reg_ax is None):
            f, (pre_reg_ax, post_reg_ax) = plt.subplots(1, 2)
        im_slice = checker_arr[slice_dx, :, :]
        pre_reg_ax.imshow(im_slice, cmap='gray', vmin=vmin, vmax=vmax)
        im_slice = checker_arr_reg[slice_dx, :, :]
        post_reg_ax.imshow(im_slice, cmap='gray', vmin=vmin, vmax=vmax)
        for ax in [pre_reg_ax, post_reg_ax]:
            ax.axis('off')
        if title is not None:
            pre_reg_ax.set_title("%s un-registered" % title)
            post_reg_ax.set_title("%s registered" % title)
        #plt.pause(0.01)


    def __get_registered_mask(self, src_mask, target_im):
        """
        Warp and resample a template mask onto the target image.

        Args:
            src_mask (SimpleITK.Image): a template mask to warp and resample
            target_im (SimpleITK.Image): a target image to specify the resampling grid

        Returns:
            SimpleITK.Image: the template mask warped and resampled into the target image space
        """
        if self.transform is None:
            mu.log("\tROIDetector::__get_registered_mask() : no transform found, need to run detect() function",
                   LogLevels.LOG_WARNING)
            return None
        return sitk.Resample(src_mask, target_im, sitk.Transform.GetInverse(self.transform), sitk.sitkNearestNeighbor)


    def __get_rotated_sampling_im(self):
        """
        Transform the template image into the target image space. This is useful for visualisation, was the template
        landmarks such as a reference T1 or T2 slice can then be sampled in the estimated/registered target space.

        Returns:
            SimpleITK.Image: The template image transformed (but not resampled) into the target image space
        """
        if self.transform is None:
            mu.log("ROIDetector::__get_rotated_sampling_im() : no transform found, need to run detect() function",
                   LogLevels.LOG_WARNING)
            return None
        # transform the reference geometry into the target space
        inv_transform = self.transform  # sitk.Transform.GetInverse(self.transform)
        # deep copy the reference image
        rotated_geom_im = sitk.GetImageFromArray(sitk.GetArrayFromImage(self.fixed_geom_im))
        rotated_geom_im.SetSpacing(self.fixed_geom_im.GetSpacing())
        rotated_geom_im.SetOrigin(self.fixed_geom_im.GetOrigin())
        rotated_geom_im.SetDirection(self.fixed_geom_im.GetDirection())
        # transform the origin
        o_new = inv_transform.TransformPoint(rotated_geom_im.GetOrigin())
        # transform the direction matrix
        d = rotated_geom_im.GetDirection()
        d_0 = [d[0], d[3], d[6]]
        d_1 = [d[1], d[4], d[7]]
        d_2 = [d[2], d[5], d[8]]
        d_0_rot = inv_transform.TransformVector(d_0, (0.0, 0.0, 0.0))
        d_1_rot = inv_transform.TransformVector(d_1, (0.0, 0.0, 0.0))
        d_2_rot = inv_transform.TransformVector(d_2, (0.0, 0.0, 0.0))
        d_new = np.zeros_like(d)
        d_new[0] = d_0_rot[0]
        d_new[1] = d_1_rot[0]
        d_new[2] = d_2_rot[0]
        d_new[3] = d_0_rot[1]
        d_new[4] = d_1_rot[1]
        d_new[5] = d_2_rot[1]
        d_new[6] = d_0_rot[2]
        d_new[7] = d_1_rot[2]
        d_new[8] = d_2_rot[2]
        # debug (remove)
        # print("origin: ", rotated_geom_im.GetOrigin())
        # print("direction: ", rotated_geom_im.GetDirection())
        # print("origin (new):", o_new)
        # print("direction (new): ", d_new)
        # re-orient and locate the image for resampling
        rotated_geom_im.SetOrigin(o_new)
        rotated_geom_im.SetDirection(d_new)
        return rotated_geom_im

    def __visualise_transformed_rois(self, mask_im,
                                     slice_dx, roi_list,
                                     ax=None,
                                     title=None):
        if self.transform is None:
            mu.log("ROIDetector::visualise_detected_T1_rois() : no transform found, need to run detect() function",
                   LogLevels.LOG_WARNING)
            return None
        rotated_geom_im = self.__get_rotated_sampling_im()
        # resample the target image and mask - onto the deformed fixed space
        # (so we can use the reference ROI slice for visualisation)
        geo_arr = sitk.GetArrayFromImage(sitk.Resample(self.target_geo_im.get_image(), rotated_geom_im,
                                                       sitk.Euler3DTransform(), sitk.sitkLinear))
        roi_arr = sitk.GetArrayFromImage(sitk.Resample(mask_im, rotated_geom_im,
                                                       sitk.Euler3DTransform(), sitk.sitkNearestNeighbor))
        self.__visualise_rois(geo_arr, roi_arr, slice_dx, roi_list, ax, title)


    def __visualise_rois(self, im_arr, roi_arr,
                         slice_dx, roi_list,
                         ax=None,
                         title=None):
        """
        Show a greyscale image and overlay coloured ROIs.

        Args:
            im_arr (numpy.array): a 3D image array
            roi_arr (numpy.array): a 3D roi array (0 as background, non-zero as regions of interest)
            slice_dx (int): an index of the 2D slice / image to display (0th dimension of the image and roi arrays)
            roi_list (list): a list of ROI indexes in the ROI array
            ax (matplotlib.axes): the axes to draw the image and ROI overlay
            title (string): a title to label the axes
        """
        if ax is None:
            f, ax = plt.subplots(1, 1)
        roi_slice = roi_arr[slice_dx, :, :]
        im_slice = im_arr[slice_dx, :, :]
        ax.imshow(im_slice, cmap='gray',
                  vmin=np.mean(im_slice)-1.0*np.std(im_slice),
                  vmax=np.mean(im_slice)+2.0*np.std(im_slice))
        i = ax.imshow(np.ma.masked_where(roi_slice == 0, roi_slice),
                      cmap='nipy_spectral', vmin=np.min(roi_list)-1, vmax=np.max(roi_list)+1,
                      interpolation='none',
                      alpha=0.7)
        ax.axis('off')
        ticks = list(range(np.min(roi_list), np.max(roi_list)+1))
        ticklabels = [ROI_IDX_LABEL_MAP[x] for x in ticks]
        cb = plt.colorbar(mappable=i, ax=ax,
                     ticks=ticks)
        cb.set_ticklabels(ticklabels=ticklabels)
        if title is not None:
            ax.set_title(title)
        #plt.pause(0.01)




class RegistrationMethodAbstract(ABC):
    def __init__(self, fixed_image, moving_image):
        self.fixed_im = fixed_image
        self.moving_im = moving_image

    @abstractmethod
    def register(self):
        return None, None

    @staticmethod
    def generate_registration_instance(reg_method, fixed_geom_im, target_geo_im):
        rego = None
        if reg_method == RegistrationOptions.COREL_GRADIENTDESCENT:
            rego = RegistrationCorrelationGradientDescent(fixed_geom_im, target_geo_im)
        elif reg_method == RegistrationOptions.MSME_GRIDSEARCH:
            rego = RegistrationMSMEGridSearch(fixed_geom_im, target_geo_im)
        elif reg_method == RegistrationOptions.TWOSTAGE_MSMEGS_CORELGD:
            rego = RegistrationTwoStage(fixed_geom_im, target_geo_im,
                                        RegistrationOptions.MSME_GRIDSEARCH,
                                        RegistrationOptions.COREL_GRADIENTDESCENT)
        assert rego is not None, "RegistrationMethodAbstract::generate_registration_instance(): " \
                                 "invalid registration method, %s" % str(reg_method)
        return rego

    def print_iteration_info(self, method):
        mu.log("\t\t{0:3} = {1:10.5f} : {2}".format(method.GetOptimizerIteration(),
                                                    method.GetMetricValue(),
                                                    method.GetOptimizerPosition()),
               LogLevels.LOG_INFO)



class RegistrationTwoStage(RegistrationMethodAbstract):
    def __init__(self, fixed_image, moving_image,
                 stage_a_registration_method=RegistrationOptions.MSME_GRIDSEARCH,
                 stage_b_registration_method=RegistrationOptions.COREL_GRADIENTDESCENT):
        super().__init__(fixed_image, moving_image)
        mu.log("RegistrationTwoStage::init()", LogLevels.LOG_INFO)
        self.stage_a_registration_method = stage_a_registration_method
        self.stage_b_registration_method = stage_b_registration_method
        self.stage_a_rego = RegistrationMethodAbstract.generate_registration_instance(stage_a_registration_method,
                                                                                      fixed_image, moving_image)
        self.stage_b_rego = RegistrationMethodAbstract.generate_registration_instance(stage_b_registration_method,
                                                                                      fixed_image, moving_image)

    def register(self):
        # stage A registration
        trans_a, metric_a = self.stage_a_rego.register()
        trans_image_a = sitk.Resample(self.moving_im, self.fixed_im, trans_a, sitk.sitkLinear)
        # stage B registration
        self.stage_b_rego = RegistrationMethodAbstract.generate_registration_instance(self.stage_b_registration_method,
                                                                                      self.fixed_im, trans_image_a)
        trans_b, metric_b = self.stage_b_rego.register()
        # compose the final transform
        combined_transform = None
        if hasattr(sitk, "CompositeTransform"):
            combined_transform = sitk.CompositeTransform(trans_a)
        elif hasattr(sitk, "Transform"):
            combined_transform = sitk.Transform(trans_a)
        else:
            mu.log("RegistrationTwoStage::register(): unable to locate simpleITK multi-stage transform", LogLevels.LOG_ERROR)
            assert False
        combined_transform.AddTransform(trans_b)
        return combined_transform, metric_b



class RegistrationCorrelationGradientDescent(RegistrationMethodAbstract):
    def __init__(self, fixed_image, moving_image):
        super().__init__(fixed_image, moving_image)
        mu.log("RegistrationCorrelationGradientDescent::init()", LogLevels.LOG_INFO)

    def register(self):
        moving = sitk.Cast(self.moving_im, sitk.sitkFloat32)
        fixed = sitk.Cast(self.fixed_im, sitk.sitkFloat32)
        # setup the registration method
        R = sitk.ImageRegistrationMethod()
        R.SetMetricAsCorrelation()
        R.SetMetricSamplingPercentage(0.01)
        R.SetMetricSamplingStrategy(R.RANDOM)
        R.SetOptimizerAsRegularStepGradientDescent(learningRate=1.0,
                                                   numberOfIterations=1000,
                                                   minStep=0.000001)
        R.SetOptimizerScalesFromPhysicalShift()
        # centre as initial transform
        initial_transform = sitk.CenteredTransformInitializer(fixed, moving, sitk.Euler3DTransform(),
                                                              sitk.CenteredTransformInitializerFilter.GEOMETRY)
        R.SetInitialTransform(initial_transform)
        R.SetInterpolator(sitk.sitkLinear)
        # hook up a logger
        R.AddCommand(sitk.sitkIterationEvent, lambda: self.print_iteration_info(R))
        # register
        final_transform = R.Execute(fixed, moving)
        return final_transform, R.GetMetricValue()


class RegistrationMSMEGridSearch(RegistrationMethodAbstract):
    def __init__(self, fixed_image, moving_image):
        super().__init__(fixed_image, moving_image)
        mu.log("RegistrationMSMEGridSearch::init()", LogLevels.LOG_INFO)

    def register(self):
        moving = sitk.Cast(self.moving_im, sitk.sitkFloat32)
        fixed = sitk.Cast(self.fixed_im, sitk.sitkFloat32)
        # setup the registration method
        R = sitk.ImageRegistrationMethod()
        R.SetMetricAsMeanSquares()
        R.SetMetricSamplingPercentage(0.001)
        R.SetMetricSamplingStrategy(R.REGULAR)
        # Number of samples for each rotational axis, 360 divided by this number gives you the angle 'incremented'
        sample_per_axis = 24
        # Note: Order of parameters is ( x-rotation, y-rotation, z-rotation, x, y, z)
        R.SetOptimizerAsExhaustive([sample_per_axis // 2, sample_per_axis // 2, sample_per_axis // 4,
                                    0, 0, 0])
        R.SetOptimizerScales(
            [2.0 * np.pi / sample_per_axis, 2.0 * np.pi / sample_per_axis, 2.0 * np.pi / sample_per_axis,
             1.0, 1.0, 1.0])
        # centre as initial transform
        initial_transform = sitk.CenteredTransformInitializer(fixed, moving, sitk.Euler3DTransform(),
                                                              sitk.CenteredTransformInitializerFilter.GEOMETRY)
        R.SetInitialTransform(initial_transform)
        R.SetInterpolator(sitk.sitkLinear)
        # hook up a logger (too much output and won't show convergance as it is a grid search)
        # R.AddCommand(sitk.sitkIterationEvent, lambda: self.print_iteration_info(R))
        # register
        final_transform = R.Execute(fixed, moving)
        # output the best transform of the grid search
        mu.log("\t\t{0:3} = {1:10.5f} : {2}".format(R.GetOptimizerIteration(),
                                                    R.GetMetricValue(),
                                                    final_transform),
               LogLevels.LOG_INFO)
        return final_transform, R.GetMetricValue()







if __name__ == "__main__":
    main()



"""
Copyright (c) 2021 James Korte, Zachary Chin

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
following conditions are met:
1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer
   in the documentation and/or other materials provided with the distribution.
3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

--------------------------------------------------------------------------------
Change Log:
--------------------------------------------------------------------------------
02-August-2021  :               (James Korte) : Initial code for MR-BIAS v0.0
  23-June-2022  :               (James Korte) : GitHub Release   MR-BIAS v1.0
"""

import re
import os
from collections import OrderedDict
from enum import IntEnum

# to handle relative paths (for calls inside or outside package)
from pathlib import Path

import SimpleITK as sitk
import numpy as np

import matplotlib as mpl

import datetime

from reportlab.lib.pagesizes import A4, A3, landscape
from reportlab.lib.units import inch, cm
import PIL
import io


MRBIAS_VERSION_NUMBER = "1.0.0"
MRBIAS_VERSION_DATE   = "13th June 2022"
MRBIAS_URL            = "http://github.com/JamesCKorte/mrbias"
MRBIAS_DOI_URL        = "http://MRBIAS_DOI_URL_TO_BE_DEFINED"

MRBIAS_REPORT_IMAGE_DPI = 500

# Want a unique ROI label (across T1/T2/PD to allow combined/mixed ROI type dataframes)
# The Label->IDX maps are then for each specific ROI type (T1/T2/PD) and also map to the input yaml files
ROI_IDX_LABEL_MAP = OrderedDict()
T1_ROI_LABEL_IDX_MAP = OrderedDict()
for i in range(14):
    ROI_IDX_LABEL_MAP[i+1] = "t1_roi_%d" % (i+1)
    T1_ROI_LABEL_IDX_MAP["t1_roi_%d" % (i+1)] = i+1
T2_ROI_LABEL_IDX_MAP = OrderedDict()
for i in range(14):
    ROI_IDX_LABEL_MAP[i+14+1] = "t2_roi_%d" % (i+1)
    T2_ROI_LABEL_IDX_MAP["t2_roi_%d" % (i+1)] = i+14+1
# create a reverse lookup
ROI_LABEL_IDX_MAP = {v: k for k, v in ROI_IDX_LABEL_MAP.items()}


class ColourSettings(object):
    def __init__(self, matplotlib_cmap_name='nipy_spectral'):
        self.cmap_name = matplotlib_cmap_name
        self.cmap = mpl.cm.get_cmap(self.cmap_name)
        t1_roi_dx_list = list(T1_ROI_LABEL_IDX_MAP.values())
        self.norm_t1 = mpl.colors.Normalize(vmin=np.min(t1_roi_dx_list)-1, vmax=np.max(t1_roi_dx_list)+1)
        t2_roi_dx_list = list(T2_ROI_LABEL_IDX_MAP.values())
        self.norm_t2 = mpl.colors.Normalize(vmin=np.min(t2_roi_dx_list)-1, vmax=np.max(t2_roi_dx_list)+1)

    def get_ROI_colour(self, roi_label):
        if roi_label in T1_ROI_LABEL_IDX_MAP.keys():
            return self.cmap(self.norm_t1(T1_ROI_LABEL_IDX_MAP[roi_label]))
        elif roi_label in T2_ROI_LABEL_IDX_MAP.keys():
            return self.cmap(self.norm_t2(T2_ROI_LABEL_IDX_MAP[roi_label]))
        else:
            log("ColourSettings::get_ROI_colour(): label %s not found returning default black" % roi_label,
                LogLevels.LOG_WARNING)
            return (0.0, 0.0, 0.0)

SEABORN_STYLE = "whitegrid"

class PDFSettings(object):
    def __init__(self):
        self.page_size = A3
        self.page_width, self.page_height = landscape(self.page_size)
        self.top_margin = 0.3 * inch
        self.left_margin = 0.3 * inch
        self.line_width = 0.2 * inch
        self.small_line_width = 0.15 * inch
        self.font_name = 'Courier' # set to a fixed width font
        self.font_size = 14
        self.small_font_size = 11


BLUE = np.array([29, 144, 170])/255.0
GREEN = np.array([140, 187, 78])/255.0
PURPLE = np.array([121, 88, 146])/255.0
RED = np.array([196, 68, 64])/255.0
ORANGE = np.array([243, 133, 51])/255.0
GREY = np.array([145, 145, 145])/255.0
BLACK = np.array([0.0, 0.0, 0.0])
WHITE = np.array([1.0, 1.0, 1.0])

key_fmt = "%s_%03d"

class LogLevels(IntEnum):
    LOG_ERROR = 1
    LOG_WARNING = 2
    LOG_INFO = 3
LOG_LEVEL_LIMIT = LogLevels.LOG_INFO
LOG_LEVEL_STR_DICT = {LogLevels.LOG_ERROR : "ERROR",
                      LogLevels.LOG_WARNING : "WARNING",
                      LogLevels.LOG_INFO : "INFO"}

class Singleton(type):
    _instances = {}
    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]

class Logger(object):#(metaclass=Singleton): # todo: change back to Singleton and create a re-routing function (call from initialise_logger)
    def __init__(self, filename=None, force_overwrite=False, write_to_screen=False):
        self.filename = filename
        self.file = None
        self.write_to_file = False
        self.write_to_screen = write_to_screen
        # set it all up
        if self.filename is not None:
            # check its not an existing file
            if os.path.isfile(self.filename) and (not force_overwrite):
                assert False, "misc_utils::Logger::__init__(): log file already exists will not overwrite : %s" % self.filename
            else:
                self.file = open(self.filename, "w")
                self.write_to_file = True

    def write(self, msg):
        # write to file if log file available
        if self.write_to_file:
            if self.file is not None:
                timestamp = datetime.datetime.now().strftime("%d.%b %Y %H:%M:%S")
                self.file.write("%s %s\n" % (timestamp, msg))
        # write to screen as well if requested
        if self.write_to_screen:
            print(msg)

    def __del__(self):
        if self.file is not None:
            self.file.close()

logger = None
def initialise_logger(filename=None, force_overwrite=False, write_to_screen=False):
    global logger
    logger = Logger(filename, force_overwrite, write_to_screen)
def detatch_logger():
    global logger
    logger = None

def log(str, log_level, log_limit=LOG_LEVEL_LIMIT):
    if log_level <= log_limit:
        msg = "[%s] %s" % (LOG_LEVEL_STR_DICT[log_level], str)
        if logger is None:
            print(msg)
        else:
            logger.write(msg)



BASE_PATH = Path(__file__).parent
# should now work within module and outside module
def reference_data_directory():
    return (BASE_PATH / os.path.join("..", "data")).resolve()
def reference_template_directory():
    return (BASE_PATH / "roi_detection_templates").resolve()
def reference_phantom_values_directory():
    return (BASE_PATH / "roi_reference_values").resolve()



def natural_sort(ll):
    convert = lambda text: int(text) if text.isdigit() else text.lower()
    alphanum_key = lambda key: [convert(c) for c in re.split('([0-9]+)', key)]
    return sorted(ll, key=alphanum_key)

# Scan a dicom directory and return a simpleITK image
def get_sitk_image_from_dicom_image_folder(dcm_dir, if_multiple_take_first=False):
    reader = sitk.ImageSeriesReader()
    series_ids = reader.GetGDCMSeriesIDs(dcm_dir)
    assert len(series_ids) > 0, "mu::get_sitk_image_from_dicom_image_folder() : no dicom series found in %s" % dcm_dir
    if len(series_ids) > 1 and (not if_multiple_take_first):
        assert False, "mu::get_sitk_image_from_dicom_image_folder() : not able to handle multiple DCM series in %s" % dcm_dir
    series_id = series_ids[0]
    dcm_filenames = reader.GetGDCMSeriesFileNames(dcm_dir, series_id)
    assert len(dcm_filenames) > 0, "mu::get_sitk_image_from_dicom_image_folder() : no dicom files found in %s" % dcm_dir
    reader.SetFileNames(dcm_filenames)
    return reader.Execute()

def safe_dir_create(d, log_prefix=""):
    if os.path.isdir(d):
        log("%s folder already exists: %s" % (log_prefix, d), LogLevels.LOG_WARNING)
        return False
    else:
        os.mkdir(d)
    return True


# converserion of matplotlib figure to a PIL for output by PDFReport
def mplcanvas_to_pil(f):
    buf = io.BytesIO()
    f.savefig(buf, format='png', dpi=MRBIAS_REPORT_IMAGE_DPI)
    buf.seek(0)
    im = PIL.Image.open(buf)
    return im

def isclose(a, b, abs_tol):
    return np.isclose(a, b, atol=abs_tol)

def find_nearest_float(value, array):
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return array[idx], abs(array[idx]-value)


"""
Copyright (c) 2021 James Korte, Zachary Chin

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
following conditions are met:
1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer
   in the documentation and/or other materials provided with the distribution.
3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

--------------------------------------------------------------------------------
Change Log:
--------------------------------------------------------------------------------
02-August-2021  :               (James Korte) : Initial code for MR-BIAS v0.0
  23-June-2022  :               (James Korte) : GitHub Release   MR-BIAS v1.0
"""

from abc import ABC, abstractmethod
from collections import OrderedDict

import SimpleITK as sitk
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from reportlab.lib.utils import ImageReader

# Code to handle running each module as a test case (from within the module)
from pathlib import Path
import sys
file = Path(__file__).resolve()
parent, root = file.parent, file.parents[1]
if str(root) not in sys.path:
    sys.path.insert(1, str(root))
# import required mrbias modules
from mrbias import misc_utils as mu
from mrbias.misc_utils import LogLevels
from mrbias.misc_utils import ROI_IDX_LABEL_MAP, T1_ROI_LABEL_IDX_MAP, T2_ROI_LABEL_IDX_MAP


def check_image_same_grid(im_a, im_b):
    spacing_check = (im_a.GetSpacing() == im_b.GetSpacing())
    origin_check = (im_a.GetOrigin() == im_b.GetOrigin())
    direction_check = (im_a.GetDirection() == im_b.GetDirection())
    mu.log("image_sets::check_image_same_grid() [Spacing:%s, Origin:%s, Direct:%s]" %
           (spacing_check, origin_check, direction_check), LogLevels.LOG_INFO)
    return spacing_check and origin_check and direction_check


# a class to bundle together multiple images
# their associated variable of interest (from DICOM tags)
# and label names and units
# also linked to a geometry image
class ImageSetAbstract(ABC):
    def __init__(self,
                 set_label,
                 sitk_im_list,
                 measurement_variable_list,
                 measurement_variable_name,
                 measurement_variable_units,
                 repetition_time_list,
                 geometry_image=None,
                 series_instance_UIDs=None,
                 bits_allocated=None, bits_stored=None, rescale_slope=None, rescale_intercept=None,
                 scanner_make=None, scanner_model=None, scanner_serial_number=None, scanner_field_strength=None,
                 date_acquired=None, time_acquired=None):
        self.label = set_label
        self.image_list = sitk_im_list
        self.meas_var_list = measurement_variable_list
        self.meas_var_name = measurement_variable_name
        self.meas_var_units = measurement_variable_units
        self.repetition_time_list = repetition_time_list
        self.geometry_image = geometry_image
        self.series_instance_UIDs = series_instance_UIDs
        self.roi_image = None
        # test if all the images in the set are in the same physical space
        assert len(self.image_list) > 0, "ImageSetAbstract[%s]: has an empty image list" % self.label
        ref_im = self.image_list[0]
        all_same_grid = True
        for meas_dx, im in enumerate(self.image_list):
            on_same_grid = check_image_same_grid(ref_im, im)
            if not on_same_grid:
                mu.log("ImageSetAbstract[%s]: image with %s=%s is on a different spatial grid"
                       % (self.label, self.meas_var_name, self.meas_var_list[meas_dx]), LogLevels.LOG_WARNING)
            all_same_grid = all_same_grid and on_same_grid
        assert all_same_grid,  "ImageSetAbstract[%s]: not all images in the set are in the same physical space!" % self.label
        # MRI scanner details
        self.scanner_make = scanner_make
        self.scanner_model = scanner_model
        self.scanner_serial_number = scanner_serial_number
        self.scanner_field_strength = scanner_field_strength
        # Time/date
        self.date = date_acquired
        self.time = time_acquired

    def __str__(self):
        r_str = "ImageSetAbstract [%s : %s]" % (self.label, type(self))
        for im, x, suid in zip(self.image_list, self.meas_var_list, self.series_instance_UIDs):
            r_str = "%s\n\t\t\t\t\t%s=%s %s : (image=[size=%s%s], SeriesUID=%s)" % (r_str,
                                                                           self.meas_var_name, x, self.meas_var_units,
                                                                           str(sitk.GetArrayFromImage(im).shape), type(im), suid)
        # and the reference geometry
        if self.geometry_image is not None:
            r_str = "%s\n\t\t\tReferenced Geometry: %s" % (r_str, str(self.geometry_image))
        return r_str

    def get_ROI_data(self):
        roi_dict = OrderedDict()
        if self.roi_image is None:
            mu.log("ImageSetAbstract::get_ROI_data() - no mask available", LogLevels.LOG_WARNING)
            return roi_dict
        # loop over rois in the mask and extract voxel data
        mask_arr = sitk.GetArrayFromImage(self.roi_image)
        mu.log("ImageSetAbstract:: get_ROI_data(): MASK values: [%d, %d]" %
               (np.min(mask_arr.flatten()), np.max(mask_arr.flatten())), LogLevels.LOG_INFO)
        mask_vals = np.unique(mask_arr.flatten())
        mask_vals = np.delete(mask_vals, np.where(mask_vals == 0)) # remove background
        n_meas = (len(self.meas_var_list))
        n_images = (len(self.image_list))
        assert n_meas == n_images, "ImageSetAbstract::get_ROI_data() - number of measurements (%d) and " \
                                   "images (%d) does not match" % (n_meas, n_images)
        for mask_val in mask_vals:
            if mask_val in mu.ROI_IDX_LABEL_MAP.keys():
                # Gets a boolean mask by comparing the mask np array with a given ROI number
                roi_boolean_mask = (mask_arr == mask_val)
                roi_mask_xyz = np.nonzero(roi_boolean_mask)
                # get the data type
                assert len(self.image_list) > 0, "ImageSetAbstract::get_ROI_data(): image_list has no images"
                image_array = sitk.GetArrayFromImage(self.image_list[0])
                dtype = type(image_array.flatten()[0])
                # iterate over the measurements/images and extract ROI voxel data
                voxel_data = np.zeros((n_meas, np.count_nonzero(roi_boolean_mask)), dtype=dtype)
                for meas_dx, image in enumerate(self.image_list):
                    image_array = sitk.GetArrayFromImage(image)
                    roi_im_data = image_array[roi_boolean_mask]
                    #print("get_ROI_data(): data type : %s" % type(roi_im_data.flatten()[0]))
                    voxel_data[meas_dx] = roi_im_data.flatten()
                # instantiate a ROI object and append to return structure
                im_set_roi = ImageSetROI(label=mu.ROI_IDX_LABEL_MAP[mask_val],
                                         voxel_data_array=voxel_data.transpose(),
                                         voxel_data_xyz=roi_mask_xyz,
                                         measurement_variable_vector=self.meas_var_list,
                                         measurement_variable_name=self.meas_var_name,
                                         measurement_variable_units=self.meas_var_units)
                roi_dict[mask_val] = im_set_roi
            else:
                mu.log("ImageSetAbstract::get_ROI_data() - skipping unknown roi_idx (%s)" % mask_val,
                       LogLevels.LOG_WARNING)
        return roi_dict


    @abstractmethod
    def update_ROI_mask(self):
        return None

    def get_set_label(self):
        return self.label

    def get_images(self):
        return self.image_list

    def get_geometry_image(self):
        return self.geometry_image

    def get_roi_image(self):
        return self.roi_image

    def set_roi_image(self, roi_sitk_image):
        mu.log("ImageSetAbstract::set_roi_image(): check ROI mask matched the ImageSet spatial grid: %s" %
               check_image_same_grid(roi_sitk_image, self.image_list[0]), LogLevels.LOG_INFO)
        self.roi_image = roi_sitk_image

    def get_measurement_variables(self):
        return self.meas_var_list

    def get_label(self):
        return self.meas_var_name

    def get_label_units(self):
        return self.meas_var_units

    def write_roi_pdf_page(self, c, sup_title=None):
        pdf = mu.PDFSettings()
        c.setFont(pdf.font_name, pdf.small_font_size)  # set to a fixed width font
        if sup_title is None:
            sup_title = "IMAGESET [%s]" % (self.get_set_label())
        # get some data from the image set
        image_list = self.get_images()
        base_im = image_list[0]
        im_spacing = base_im.GetSpacing()
        base_arr = sitk.GetArrayFromImage(base_im)
        mask_im = self.get_roi_image()
        mask_arr = sitk.GetArrayFromImage(mask_im)

        roi_vals = np.unique(mask_arr.flatten())[1:]

        # draw the ROI location summary figure
        # -----------------------------------------------------------
        # setup figure
        n_rois = len(roi_vals)
        n_rows = 2
        rois_per_row = int(np.ceil(n_rois / float(n_rows)))
        # f, axes_arr = plt.subplots(n_rows*2, rois_per_row)
        f = plt.figure()  # constrained_layout=True)
        f.suptitle(sup_title)
        f.set_size_inches(14, 8)
        # setup the layout
        gs0 = gridspec.GridSpec(2, 2, figure=f,
                                bottom=0.1, top=0.85, left=0.1, right=0.9,
                                width_ratios=[1, 2], hspace=0.5)

        gs00 = gridspec.GridSpecFromSubplotSpec(1, 1, subplot_spec=gs0[:, 0])
        gs01 = gridspec.GridSpecFromSubplotSpec(2, rois_per_row, subplot_spec=gs0[0, 1],
                                                hspace=0.01, wspace=0.01)
        gs02 = gridspec.GridSpecFromSubplotSpec(2, rois_per_row, subplot_spec=gs0[1, 1],
                                                hspace=0.01, wspace=0.01)
        # make the main image axes
        ax_glob = f.add_subplot(gs00[:, :])
        # and for the zoom axes
        ax_row_1a = [f.add_subplot(gs01[0, x]) for x in range(rois_per_row)]
        ax_row_1b = [f.add_subplot(gs01[1, x]) for x in range(rois_per_row)]
        ax_row_2a = [f.add_subplot(gs02[0, x]) for x in range(rois_per_row)]
        ax_row_2b = [f.add_subplot(gs02[1, x]) for x in range(rois_per_row)]

        # draw the main slice for reference
        roi_xyz = np.nonzero(mask_arr == roi_vals[0])
        c_x, c_y, c_z = np.median(roi_xyz[0]).astype(int), \
                        np.median(roi_xyz[1]).astype(int), \
                        np.median(roi_xyz[2]).astype(int)
        roi_slice = mask_arr[c_x, :, :]
        im_slice = base_arr[c_x, :, :]
        ax_glob.imshow(im_slice, cmap='gray')
        i = ax_glob.imshow(np.ma.masked_where(roi_slice == 0, roi_slice),
                           cmap='nipy_spectral', vmin=np.min(roi_vals) - 1, vmax=np.max(roi_vals) + 1,
                           interpolation='none',
                           alpha=0.7)
        ax_glob.axis('off')
        ticks = list(range(np.min(roi_vals), np.max(roi_vals) + 1))
        ticklabels = [ROI_IDX_LABEL_MAP[x] for x in ticks]
        cb = plt.colorbar(mappable=i, ax=ax_glob,
                          ticks=ticks)
        cb.set_ticklabels(ticklabels=ticklabels)

        # loop over the axes and plot the individual roi regions
        roi_dx = 0
        for ax_row_dx, axes_row in zip([0, 2],
                                       [ax_row_1a, ax_row_2a]):
            for ax_dx, ax in enumerate(axes_row):
                if roi_dx < n_rois:
                    roi_xyz = np.nonzero(mask_arr == roi_vals[roi_dx])
                    c_x, c_y, c_z = np.median(roi_xyz[0]).astype(int), \
                                    np.median(roi_xyz[1]).astype(int), \
                                    np.median(roi_xyz[2]).astype(int)
                    c_x_min, c_y_min, c_z_min = np.min(roi_xyz[0]).astype(int), \
                                                np.min(roi_xyz[1]).astype(int), \
                                                np.min(roi_xyz[2]).astype(int)
                    c_x_max, c_y_max, c_z_max = np.max(roi_xyz[0]).astype(int), \
                                                np.max(roi_xyz[1]).astype(int), \
                                                np.max(roi_xyz[2]).astype(int)
                    padd_yz = int(np.round(10. / im_spacing[0]))
                    padd_x = int(np.round(10. / im_spacing[2]))
                    extent_x_a = c_x_min - padd_x if (c_x_min - padd_x) >= 0 else 0
                    extent_x_b = c_x_max + padd_x if (c_x_max + padd_x) <= base_arr.shape[0] else base_arr.shape[0]
                    extent_y_a = c_y_min - padd_yz if (c_y_min - padd_yz) >= 0 else 0
                    extent_y_b = c_y_max + padd_yz if (c_y_max + padd_yz) <= base_arr.shape[1] else base_arr.shape[1]
                    extent_z_a = c_z_min - padd_yz if (c_z_min - padd_yz) >= 0 else 0
                    extent_z_b = c_z_max + padd_yz if (c_z_max + padd_yz) <= base_arr.shape[2] else base_arr.shape[2]

                    zoom_arr = base_arr[c_x, extent_y_a:extent_y_b, extent_z_a:extent_z_b]
                    zoom_extent = [0, zoom_arr.shape[0] * im_spacing[1],
                                   0, zoom_arr.shape[1] * im_spacing[0]]
                    roi_slice = mask_arr[c_x, extent_y_a:extent_y_b, extent_z_a:extent_z_b]
                    ax.imshow(zoom_arr,
                              extent=zoom_extent,
                              cmap='gray')
                    ax.imshow(np.ma.masked_where(roi_slice == 0, roi_slice),
                              extent=zoom_extent,
                              cmap='nipy_spectral', vmin=np.min(roi_vals) - 1, vmax=np.max(roi_vals) + 1,
                              interpolation='none',
                              alpha=0.7)
                    ax.set_xticks([])
                    ax.set_xticklabels([])
                    ax.set_yticks([])
                    ax.set_yticklabels([])
                    ax.set_title("ROI_%d\n(n=%d)" % (roi_dx + 1, np.count_nonzero(mask_arr == roi_vals[roi_dx])))
                    if ax_dx == 0:
                        ax.set_ylabel("axial")

                    # if its a 3D image try again
                    sag_ax = None
                    if ax_row_dx == 0:
                        sag_ax = ax_row_1b[ax_dx]
                    if ax_row_dx == 2:
                        sag_ax = ax_row_2b[ax_dx]
                    if sag_ax is not None:
                        if (c_x_max - c_x_min) > 1:
                            zoom_arr = base_arr[extent_x_a:extent_x_b, c_y, extent_z_a:extent_z_b]
                            zoom_extent = [0, zoom_arr.shape[1] * im_spacing[0],
                                           0, zoom_arr.shape[0] * im_spacing[2]]
                            sag_ax.imshow(zoom_arr,
                                          extent=zoom_extent,
                                          cmap='gray')
                            roi_slice = mask_arr[extent_x_a:extent_x_b, c_y, extent_z_a:extent_z_b]
                            sag_ax.imshow(np.ma.masked_where(roi_slice == 0, roi_slice),
                                          extent=zoom_extent,
                                          cmap='nipy_spectral', vmin=np.min(roi_vals) - 1, vmax=np.max(roi_vals) + 1,
                                          interpolation='none',
                                          alpha=0.7)
                            if ax_dx == 0:
                                sag_ax.set_ylabel("sagittal")

                            sag_ax.set_xticks([])
                            sag_ax.set_xticklabels([])
                            sag_ax.set_yticks([])
                            sag_ax.set_yticklabels([])
                        else:
                            # hide any unused axes
                            sag_ax.axis('off')

                    roi_dx = roi_dx + 1
                else:
                    # hide any unused axes
                    ax.axis('off')
        # draw it on the pdf
        pil_f = mu.mplcanvas_to_pil(f)
        width, height = pil_f.size
        height_3d, width_3d = pdf.page_width * (height / width), pdf.page_width
        c.drawImage(ImageReader(pil_f),
                    0,
                    pdf.page_height - pdf.top_margin - height_3d - pdf.line_width,
                    width_3d,
                    height_3d)
        plt.close(f)
        # -------------------------------------------------------------
        c.showPage()  # new page


class ImageSetROI(object):
    def __init__(self, label,
                 voxel_data_array, voxel_data_xyz, measurement_variable_vector,
                 measurement_variable_name, measurement_variable_units
                 ):
        mu.log("\t\t\tImageSetROI::__init__(): creating %s with %d voxels %s (each with %d measurements)" %
               (label, voxel_data_array.shape[0], type(voxel_data_array.flatten()[0]), voxel_data_array.shape[1]),
               LogLevels.LOG_INFO)
        self.label = label
        self.voxel_data_array = voxel_data_array
        self.voxel_data_xyz = voxel_data_xyz
        self.meas_var_vector = measurement_variable_vector
        self.meas_var_name = measurement_variable_name
        self.meas_var_units = measurement_variable_units


class ImageBasic(object):
    def __init__(self,
                 label, sitk_im,
                 series_instance_UID=None,
                 bits_allocated=None,
                 bits_stored=None,
                 rescale_slope=None,
                 rescale_intercept=None):
        self.label = label
        self.im = sitk_im
        self.series_instance_UID = series_instance_UID
        # image data information
        self.bits_allocated = bits_allocated
        self.bits_stored = bits_stored
        self.rescale_slope = rescale_slope
        self.rescale_intercept = rescale_intercept
    def get_image(self):
        return self.im
    def get_label(self):
        return self.label
    def __str__(self):
        return "%s (image=[size%s%s], SeriesUID=%s)" % (self.label, str(sitk.GetArrayFromImage(self.im).shape),
                                                        type(self.im), self.series_instance_UID)


class ImageGeometric(ImageBasic):
    def __init__(self,
                 label, sitk_im,
                 series_instance_UID=None,
                 bits_allocated=None,
                 bits_stored=None,
                 rescale_slope=None,
                 rescale_intercept=None):
        super().__init__(label, sitk_im, series_instance_UID,
                         bits_allocated, bits_stored, rescale_slope, rescale_intercept)
        self.roi_mask_image_PD = None
        self.roi_mask_image_T1 = None
        self.roi_mask_image_T2 = None
    def set_proton_density_mask(self, mask_sitk_image):
        assert isinstance(mask_sitk_image, sitk.Image), "ImageGeometric::set_proton_density_mask() expects a" \
                                                        "SimpleITK image"
        self.roi_mask_image_PD = mask_sitk_image
    def set_T1_mask(self, mask_sitk_image):
        assert isinstance(mask_sitk_image, sitk.Image), "ImageGeometric::set_T1_mask() expects a" \
                                                        "SimpleITK image"
        # check the geom image and mask are on the same grid
        assert check_image_same_grid(self.get_image(), mask_sitk_image), \
            "ImageGeometric::set_T1_mask(): mask image grid does not match the geometric image!"
        mask_arr = sitk.GetArrayFromImage(mask_sitk_image)
        mu.log("ImageGeometric::set_T1_mask(): detected mask value range [%d, %d]" %
               (np.min(mask_arr), np.max(mask_arr)), LogLevels.LOG_INFO)
        self.roi_mask_image_T1 = mask_sitk_image
    def set_T2_mask(self, mask_sitk_image):
        assert isinstance(mask_sitk_image, sitk.Image), "ImageGeometric::set_T2_mask() expects a" \
                                                        "SimpleITK image"
        # check the geom image and mask are on the same grid
        assert check_image_same_grid(self.get_image(), mask_sitk_image), \
            "ImageGeometric::set_T2_mask(): mask image grid does not match the geometric image!"
        mask_arr = sitk.GetArrayFromImage(mask_sitk_image)
        mu.log("ImageGeometric::set_T2_mask(): detected mask value range [%d, %d]" %
               (np.min(mask_arr), np.max(mask_arr)), LogLevels.LOG_INFO)
        self.roi_mask_image_T2 = mask_sitk_image
    def get_T1_mask(self):
        assert self.roi_mask_image_T1 is not None, "ImageGeometric::get_T1_mask(): no mask available"
        return self.roi_mask_image_T1
    def get_T2_mask(self):
        assert self.roi_mask_image_T2 is not None, "ImageGeometric::get_T2_mask(): no mask available"
        return self.roi_mask_image_T2


class ImageProtonDensity(ImageBasic):
    def __init__(self,
                 label, sitk_im,
                 geometry_sitk_im=None,
                 series_instance_UID=None,
                 bits_allocated=None,
                 bits_stored=None,
                 rescale_slope=None,
                 rescale_intercept=None):
        super().__init__(label, sitk_im, series_instance_UID,
                         bits_allocated, bits_stored, rescale_slope, rescale_intercept)
        self.geometry_image = geometry_sitk_im
        self.roi_image = None
    def __str__(self):
        return "%s\n\t\t\t\t\tReferenced Geometry: %s" % (super().__str__(),
                                                          str(self.geometry_image))

class ImageSetT1VIR(ImageSetAbstract):
    def __init__(self,
                 set_label,
                 sitk_im_list,
                 inversion_time_list,
                 repetition_time_list,
                 geometry_image,
                 series_instance_UIDs=None,
                 bits_allocated=None,
                 bits_stored=None,
                 rescale_slope=None,
                 rescale_intercept=None,
                 scanner_make=None, scanner_model=None, scanner_serial_number=None, scanner_field_strength=None,
                 study_date=None, study_time=None):
        super().__init__(set_label=set_label,
                         sitk_im_list=sitk_im_list,
                         measurement_variable_list=inversion_time_list,
                         measurement_variable_name="inversion time",
                         repetition_time_list=repetition_time_list,
                         measurement_variable_units="ms",
                         geometry_image=geometry_image,
                         series_instance_UIDs=series_instance_UIDs,
                         bits_allocated=bits_allocated, bits_stored=bits_stored,
                         rescale_slope=rescale_slope, rescale_intercept=rescale_intercept,
                         scanner_make=scanner_make, scanner_model=scanner_model, scanner_serial_number=scanner_serial_number,
                         scanner_field_strength=scanner_field_strength, date_acquired=study_date, time_acquired=study_time)
    def get_inversion_times(self):
        return super().get_measurement_variables()
    def set_T1_roi_mask(self, roi_sitk_image):
        super().set_roi_image(roi_sitk_image)
    def update_ROI_mask(self):
        if self.geometry_image is None:
            mu.log("ImageSetT1VIR::update_ROI_mask(): no geometry image set", LogLevels.LOG_WARNING)
            return None
        if not (len(self.image_list) > 0):
            mu.log("ImageSetT1VIR::update_ROI_mask(): no images in set", LogLevels.LOG_WARNING)
            return None
        # # todo: remove print once error resolved
        # print(self.geometry_image.get_label(),
        #       self.geometry_image.roi_mask_image_T1,
        #       self.geometry_image.roi_mask_image_T2)
        mask_arr = sitk.GetArrayFromImage(self.geometry_image.get_T1_mask())
        mu.log("ImageSetT1VIR[%s]::update_ROI_mask() : orig mask values [%d, %d]" %
               (self.get_set_label(), np.min(mask_arr), np.max(mask_arr)), LogLevels.LOG_INFO)
        resampled_mask = sitk.Resample(self.geometry_image.get_T1_mask(),
                                       self.image_list[0],
                                       sitk.Euler3DTransform(), sitk.sitkNearestNeighbor)
        mask_arr = sitk.GetArrayFromImage(resampled_mask)
        mu.log("ImageSetT1VIR[%s]::update_ROI_mask() : resampled mask values [%d, %d]" %
               (self.get_set_label(), np.min(mask_arr), np.max(mask_arr)), LogLevels.LOG_INFO)
        self.set_T1_roi_mask(resampled_mask)



class ImageSetT1VFA(ImageSetAbstract):
    def __init__(self,
                 set_label,
                 sitk_im_list,
                 flip_angle_list,
                 repetition_time_list,
                 geometry_image,
                 series_instance_UIDs=None,
                 bits_allocated=None,
                 bits_stored=None,
                 rescale_slope=None,
                 rescale_intercept=None,
                 scanner_make=None, scanner_model=None, scanner_serial_number=None, scanner_field_strength=None,
                 study_date=None, study_time=None):
        super().__init__(set_label=set_label,
                         sitk_im_list=sitk_im_list,
                         measurement_variable_list=flip_angle_list,
                         measurement_variable_name="flip angle",
                         measurement_variable_units="deg.",
                         repetition_time_list=repetition_time_list,
                         geometry_image=geometry_image,
                         series_instance_UIDs=series_instance_UIDs,
                         bits_allocated=bits_allocated, bits_stored=bits_stored,
                         rescale_slope=rescale_slope, rescale_intercept=rescale_intercept,
                         scanner_make=scanner_make, scanner_model=scanner_model, scanner_serial_number=scanner_serial_number,
                         scanner_field_strength=scanner_field_strength, date_acquired=study_date, time_acquired=study_time)
    def get_flip_angles(self):
        return super().get_measurement_variables()
    def set_T1_roi_mask(self, roi_sitk_image):
        super().set_roi_image(roi_sitk_image)
    def update_ROI_mask(self):
        if self.geometry_image is None:
            mu.log("ImageSetT1VFA::update_ROI_mask(): no geometry image set", LogLevels.LOG_WARNING)
        if not (len(self.image_list) > 0):
            mu.log("ImageSetT1VFA::update_ROI_mask(): no images in set", LogLevels.LOG_WARNING)
            return None
        mask_arr = sitk.GetArrayFromImage(self.geometry_image.get_T1_mask())
        mu.log("ImageSetT1VFA[%s]::update_ROI_mask() : orig mask values [%d, %d]" %
               (self.get_set_label(), np.min(mask_arr), np.max(mask_arr)), LogLevels.LOG_INFO)
        resampled_mask = sitk.Resample(self.geometry_image.get_T1_mask(),
                                       self.image_list[0],
                                       sitk.Euler3DTransform(), sitk.sitkNearestNeighbor)
        mask_arr = sitk.GetArrayFromImage(resampled_mask)
        mu.log("ImageSetT1VFA[%s]::update_ROI_mask() : resampled mask values [%d, %d]" %
               (self.get_set_label(), np.min(mask_arr), np.max(mask_arr)), LogLevels.LOG_INFO)
        self.set_T1_roi_mask(resampled_mask)


class ImageSetT2MSE(ImageSetAbstract):
    def __init__(self,
                 set_label,
                 sitk_im_list,
                 echo_time_list,
                 repetition_time_list,
                 geometry_image,
                 series_instance_UIDs=None,
                 bits_allocated=None,
                 bits_stored=None,
                 rescale_slope=None,
                 rescale_intercept=None,
                 scanner_make=None, scanner_model=None, scanner_serial_number=None, scanner_field_strength=None,
                 study_date=None, study_time=None):
        super().__init__(set_label=set_label,
                         sitk_im_list=sitk_im_list,
                         measurement_variable_list=echo_time_list,
                         measurement_variable_name="echo time",
                         measurement_variable_units="ms",
                         repetition_time_list=repetition_time_list,
                         geometry_image=geometry_image,
                         series_instance_UIDs=series_instance_UIDs,
                         bits_allocated=bits_allocated, bits_stored=bits_stored,
                         rescale_slope=rescale_slope, rescale_intercept=rescale_intercept,
                         scanner_make=scanner_make, scanner_model=scanner_model, scanner_serial_number=scanner_serial_number,
                         scanner_field_strength=scanner_field_strength, date_acquired=study_date, time_acquired=study_time)
    def get_echo_times(self):
        return super().get_measurement_variables()
    def set_T2_roi_mask(self, roi_sitk_image):
        super().set_roi_image(roi_sitk_image)
    def update_ROI_mask(self):
        if self.geometry_image is None:
            mu.log("ImageSetT2MSE::update_ROI_mask(): no geometry image set", LogLevels.LOG_WARNING)
            return None
        if not (len(self.image_list) > 0):
            mu.log("ImageSetT2MSE::update_ROI_mask(): no images in set", LogLevels.LOG_WARNING)
            return None
        mask_arr = sitk.GetArrayFromImage(self.geometry_image.get_T2_mask())
        mu.log("ImageSetT2MSE[%s]::update_ROI_mask() : orig mask values [%d, %d]" %
               (self.get_set_label(), np.min(mask_arr), np.max(mask_arr)), LogLevels.LOG_INFO)
        resampled_mask = sitk.Resample(self.geometry_image.get_T2_mask(),
                                       self.image_list[0],
                                       sitk.Euler3DTransform(), sitk.sitkNearestNeighbor)
        mask_arr = sitk.GetArrayFromImage(resampled_mask)
        mu.log("ImageSetT2MSE[%s]::update_ROI_mask() : resampled mask values [%d, %d]" %
               (self.get_set_label(), np.min(mask_arr), np.max(mask_arr)), LogLevels.LOG_INFO)
        self.set_T2_roi_mask(resampled_mask)


from setuptools import setup

setup(
   name='mrbias',
   version='1.0',
   description='Magnetic Resonance BIomarker Assessment Software (MR-BIAS) is an automated tool for extracting quantitative MR parameters from NIST/ISMRM system phantom images',
   author='James Korte',
   author_email='korte.james@gmail.com',
   url='https://github.com/JamesCKorte/mrbias',
   packages=['mrbias'],  #same as name
   install_requires=['Pillow', 'lmfit', 'SimpleITK', 'numpy', 'pandas', 'scipy', 'pyyaml', 'pydicom', 'matplotlib', 'seaborn', 'reportlab'], #external packages as dependencies
)